import os.path


ORG_NAME = 'pytoutv'
APP_NAME = 'qtoutv'
DAT_DIR = 'dat'
UI_DIR = os.path.join(DAT_DIR, 'ui')
ICONS_DIR = os.path.join(DAT_DIR, 'icons')
CONTRIBUTORS = [
    'Simon Marchi',
    'Philippe Proulx',
    'Benjamin Vanheuverzwijn',
    'Alexandre Vézina',
    'Alexis Dorais-Joncas',
    'Israel Halle',
    'Marc-Étienne M. Leveillé',
    'Simon Carpentier',
]
